#include <stdlib.h>
#include <math.h>
#include <err.h>
#include "simulation.h"
#include "arrays.h"

void
hydrostatic_fluid_pressure_distribution(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->p_f_ghost[i + 1] = sim->p_f_top
		                        + sim->phi[i] * sim->rho_f * sim->G
		                        * (sim->L_z - sim->z[i]);
}

static double
diffusivity(struct simulation *sim, int i) {
	if (sim->D > 0.0)
		return sim->D;
	else
		return sim->k[i]
		       / ((sim->alpha + sim->phi[i] * sim->beta_f) * sim->mu_f);
}

/* Determines the largest time step for the current simulation state. Note
 * that the time step should be recalculated if cell sizes or
 * diffusivities (i.e., permeabilities, porosities, viscosities, or
 * compressibilities) change. The safety factor should be in ]0;1] */
int
set_largest_fluid_timestep(struct simulation *sim, const double safety)
{
	int i;
	double dx_min, diff, diff_max, *dx;

	dx = spacing(sim->z, sim->nz);
	dx_min = INFINITY;
	for (i = 0; i < sim->nz - 1; ++i) {
		if (dx[i] < 0.0) {
			fprintf(stderr, "error: cell spacing negative (%g) in cell %d\n",
			        dx[i], i);
			free(dx);
			return 1;
		}
		if (dx[i] < dx_min)
			dx_min = dx[i];
	}
	free(dx);

	diff_max = -INFINITY;
	for (i = 0; i < sim->nz; ++i) {
		diff = diffusivity(sim, i);
		if (diff > diff_max)
			diff_max = diff;
	}

	sim->dt = safety * 0.5 * dx_min * dx_min / diff_max;
	if (sim->file_dt * 0.5 < sim->dt)
	    sim->dt = sim->file_dt;

	return 0;
}

static double
sine_wave(const double time,
          const double ampl,
          const double freq,
          const double phase)
{
	return ampl * sin(2.0 * PI * freq * time + phase);
}

static double
triangular_pulse(const double time,
                 const double peak_ampl,
                 const double freq,
                 const double peak_time)
{
	if (peak_time - 1.0 / (2.0 * freq) < time && time <= peak_time)
		return peak_ampl * 2.0 * freq * (time - peak_time) + peak_ampl;
	else if (peak_time < time && time < peak_time + 1.0 / (2.0 * freq))
		return peak_ampl * 2.0 * freq * (peak_time - time) + peak_ampl;
	else
		return 0.0;
}

static double
square_pulse(const double time,
             const double peak_ampl,
             const double freq,
             const double peak_time)
{
	if (peak_time - 1.0 / (2.0 * freq) < time &&
	    time < peak_time + 1.0 / (2.0 * freq))
		return peak_ampl;
	else
		return 0.0;
}

static void
set_fluid_bcs(double *p_f_ghost, struct simulation *sim, const double p_f_top)
{
	/* correct ghost node at top BC for hydrostatic pressure gradient */
	set_bc_dirichlet(p_f_ghost, sim->nz, +1,
	                 p_f_top - sim->phi[0] * sim->rho_f * sim->G * sim->dz);
	p_f_ghost[sim->nz] = p_f_top;	/* include top node in BC */
	set_bc_neumann(p_f_ghost,
	               sim->nz,
	               -1,
	               sim->phi[0] * sim->rho_f * sim->G,
	               sim->dz);
}

static double
darcy_pressure_change_1d(const int i,
                         const int nz,
                         const double *p_f_ghost_in,
                         const double *phi,
                         const double *phi_dot,
                         const double *k,
                         const double dz,
                         const double beta_f,
                         const double alpha,
                         const double mu_f,
                         const double D)
{
	double k_, div_k_grad_p, k_zn, k_zp;

	if (D > 0.0)
		return D * (p_f_ghost_in[i + 2]
		            - 2.0 * p_f_ghost_in[i + 1]
		            + p_f_ghost_in[i]) / (dz * dz);
	else {
		k_ = k[i];
		if (i == 0)
			k_zn = k_;
		else
			k_zn = k[i - 1];
		if (i == nz - 1)
			k_zp = k_;
		else
			k_zp = k[i + 1];

		div_k_grad_p = (2.0 * k_zp * k_ / (k_zp + k_)
		                * (p_f_ghost_in[i + 2] - p_f_ghost_in[i + 1]) / dz
		                - 2.0 * k_zn * k_ / (k_zn + k_)
		                * (p_f_ghost_in[i + 1] - p_f_ghost_in[i]) / dz
) / dz;

#ifdef DEBUG
		printf("%s [%d]: phi=%g\tdiv_k_grad_p=%g\tphi_dot=%g\n",
		        __func__, i, phi[i], div_k_grad_p, phi_dot[i]);

		printf("  p[%d, %d, %d] = [%g, %g, %g]\tk=[%g, %g, %g]\n",
		       i, i + 1, i + 2,
		       p_f_ghost_in[i], p_f_ghost_in[i + 1], p_f_ghost_in[i + 2],
		       k_zn, k_, k_zp);
#endif

		return 1.0 / ((alpha + beta_f * phi[i]) * mu_f) * div_k_grad_p
		       - 1.0 / ((alpha + beta_f * phi[i]) * (1.0 - phi[i])) * phi_dot[i];
	}
}

static double
darcy_pressure_change_1d_impl(const int i,
                              const int nz,
                              const double dt,
                              const double *p_f_old_val,
                              const double *p_f_ghost_in,
                              double *p_f_ghost_out,
                              const double *phi,
                              const double *phi_dot,
                              const double *k,
                              const double dz,
                              const double beta_f,
                              const double alpha,
                              const double mu_f,
                              const double D)
{
	double k_, div_k_grad_p, k_zn, k_zp, rhs_term, omega = 1.0;

	if (D > 0.0)
		return D * (p_f_ghost_in[i + 2]
		            - 2.0 * p_f_ghost_in[i + 1]
		            + p_f_ghost_in[i]) / (dz * dz);
	else {
		k_ = k[i];
		if (i == 0)
			k_zn = k_;
		else
			k_zn = k[i - 1];
		if (i == nz - 1)
			k_zp = k_;
		else
			k_zp = k[i + 1];

		rhs_term = dt / ((alpha + beta_f * phi[i]) * mu_f)
		           * ((2.0 * k_zp * k_ / (k_zp + k_) / (dz * dz))
		              + (2.0 * k_zn * k_ / (k_zn + k_) / (dz * dz)));

		p_f_ghost_out[i + 1] = 1.0 / (1.0 + rhs_term)
		                       * (p_f_old_val[i + 1] + dt
		                       * (1.0 / ((alpha + beta_f * phi[i]) * mu_f)
		                       * (2.0 * k_zp * k_ / (k_zp + k_)
		                       * (p_f_ghost_in[i + 2]) / dz
		                       - 2.0 * k_zn * k_ / (k_zn + k_)
		                       * -p_f_ghost_in[i] / dz) / dz
		                       - 1.0 / ((alpha + beta_f * phi[i])
		                       * (1.0 - phi[i])) * phi_dot[i]));
		p_f_ghost_out[i + 1] = omega * p_f_ghost_out[i + 1]
		                       + (1.0 - omega) * p_f_ghost_in[i + 1];

		div_k_grad_p = (2.0 * k_zp * k_ / (k_zp + k_)
		                 * (p_f_ghost_out[i + 2] - p_f_ghost_out[i + 1]) / dz
		                 - 2.0 * k_zn * k_ / (k_zn + k_)
		                 * (p_f_ghost_out[i + 1] - p_f_ghost_out[i]) / dz
		                ) / dz;
#ifdef DEBUG
		printf("%s [%d]: phi=%g\tdiv_k_grad_p=%g\tphi_dot=%g\n",
		        __func__, i, phi[i], div_k_grad_p, phi_dot[i]);

		printf("  p[%d, %d, %d] = [%g, %g, %g]\tk=[%g, %g, %g]\n",
		       i, i + 1, i + 2,
		       p_f_ghost_in[i], p_f_ghost_in[i + 1], p_f_ghost_in[i + 2],
		       k_zn, k_, k_zp);
#endif
		/* use the values from the next time step as the time derivative for this iteration */
		return 1.0 / ((alpha + beta_f * phi[i]) * mu_f) * div_k_grad_p
		       - 1.0 / ((alpha + beta_f * phi[i]) * (1.0 - phi[i])) * phi_dot[i];
	}
}

int
darcy_solver_1d(struct simulation *sim,
                const int max_iter,
                const double rel_tol)
{
	int i, iter, solved = 0;
	double epsilon, p_f_top, omega, r_norm_max = NAN, *k_n, *phi_n;

	copy_values(sim->p_f_dot, sim->p_f_dot_old, sim->nz);

	/* choose integration method, parameter in [0.0; 1.0]
	 * epsilon = 0.0: explicit
	 * epsilon = 0.5: Crank-Nicolson
	 * epsilon = 1.0: implicit */
	epsilon = 0.5;

	/* underrelaxation parameter in ]0.0; 1.0],
	 * where omega = 1.0: no underrelaxation */
	omega = 1.0;

	for (i = 0; i < sim->nz; ++i)
		sim->p_f_dot_impl[i] = sim->p_f_dot_expl[i] = 0.0;

	if (isnan(sim->p_f_mod_pulse_time))
		p_f_top = sim->p_f_top
		          + sine_wave(sim->t,
		                      sim->p_f_mod_ampl,
		                      sim->p_f_mod_freq,
		                      sim->p_f_mod_phase);
	else if (sim->p_f_mod_pulse_shape == 1)
		p_f_top = sim->p_f_top
		          + square_pulse(sim->t,
		                         sim->p_f_mod_ampl,
		                         sim->p_f_mod_freq,
		                         sim->p_f_mod_pulse_time);
	else
		p_f_top = sim->p_f_top
		          + triangular_pulse(sim->t,
		                             sim->p_f_mod_ampl,
		                             sim->p_f_mod_freq,
		                             sim->p_f_mod_pulse_time);

	/* set fluid BCs (1 of 2) */
	set_fluid_bcs(sim->p_f_ghost, sim, p_f_top);
	set_fluid_bcs(sim->p_f_next_ghost, sim, p_f_top);

	/* explicit solution to pressure change */
	if (epsilon < 1.0) {
#ifdef DEBUG
		printf("\nEXPLICIT SOLVER IN %s\n", __func__);
#endif
		for (i = 0; i < sim->nz - 1; ++i)
			sim->p_f_dot_expl[i] = darcy_pressure_change_1d(i,
			                                                sim->nz,
			                                                sim->p_f_ghost,
			                                                sim->phi,
			                                                sim->phi_dot,
			                                                sim->k,
			                                                sim->dz,
			                                                sim->beta_f,
			                                                sim->alpha,
			                                                sim->mu_f,
			                                                sim->D);
	}

	/* implicit solution with Jacobian iterations */
	if (epsilon > 0.0) {

#ifdef DEBUG
		printf("\nIMPLICIT SOLVER IN %s\n", __func__);
#endif

		k_n = zeros(sim->nz);
		phi_n = zeros(sim->nz);
		if (sim->transient)
			for (i = 0; i < sim->nz; ++i) {
				/* grabbing the n + 1 iteration values for k and phi */
				phi_n[i] = sim->phi[i] + sim->dt * sim->phi_dot[i];
				k_n[i] = kozeny_carman(sim->d, phi_n[i]);
			}
		else
			for (i = 0; i < sim->nz; ++i) {
				phi_n[i] = sim->phi[i];
				k_n[i] = sim->k[i];
			}
		copy_values(sim->p_f_next_ghost, sim->tmp_ghost, sim->nz + 2);

		for (iter = 0; iter < max_iter; ++iter) {
			copy_values(sim->p_f_dot_impl, sim->fluid_old_val, sim->nz);

#ifdef DEBUG
			puts(".. p_f_ghost bfore BC:");
			print_array(sim->tmp_ghost, sim->nz + 2);
#endif

			/* set fluid BCs (2 of 2) */
			set_fluid_bcs(sim->tmp_ghost, sim, p_f_top);

#ifdef DEBUG
			puts(".. p_f_ghost after BC:");
			print_array(sim->tmp_ghost, sim->nz + 2);
#endif

			for (i = 0; i < sim->nz - 1; ++i)
				sim->p_f_dot_impl[i] = darcy_pressure_change_1d_impl(i,
				                                                     sim->nz,
				                                                     sim->dt,
				                                                     sim->p_f_ghost,
				                                                     sim->tmp_ghost,
				                                                     sim->p_f_next_ghost,
				                                                     phi_n,
				                                                     sim->phi_dot,
				                                                     k_n,
				                                                     sim->dz,
				                                                     sim->beta_f,
				                                                     sim->alpha,
				                                                     sim->mu_f,
				                                                     sim->D);

			for (i = 0; i < sim->nz; ++i)
				if (isnan(sim->p_f_dot_impl[i]))
					errx(1, "NaN at sim->p_f_dot_impl[%d] (t = %g s, iter = %d)",
						 i, sim->t, iter);

			set_fluid_bcs(sim->p_f_next_ghost, sim, p_f_top);
			for (i = 0; i < sim->nz-1; ++i)
				sim->p_f_dot_impl_r_norm[i] = fabs(residual(sim->p_f_next_ghost[i],
				                                            sim->tmp_ghost[i]));
			r_norm_max = max(sim->p_f_dot_impl_r_norm, sim->nz - 1);
			copy_values(sim->p_f_next_ghost, sim->tmp_ghost, sim->nz + 2);

#ifdef DEBUG
			puts(".. p_f_ghost_new:");
			print_array(sim->tmp_ghost, sim->nz + 2);
#endif

			if (r_norm_max <= rel_tol) {
#ifdef DEBUG
				printf(".. Iterative solution converged after %d iterations\n", iter);
#endif
				solved = 1;
				break;
			}
		}
		free(k_n);
		free(phi_n);
		if (!solved) {
			fprintf(stderr, "darcy_solver_1d: ");
			fprintf(stderr, "Solution did not converge after %d iterations\n",
			        iter);
			fprintf(stderr, ".. Residual normalized error: %f\n", r_norm_max);
		}
	} else
		solved = 1;

	for (i = 0; i < sim->nz; ++i)
		sim->p_f_dot[i] = epsilon * sim->p_f_dot_impl[i]
		                  + (1.0 - epsilon) * sim->p_f_dot_expl[i];

	for (i = 0; i < sim->nz; ++i)
		    sim->p_f_dot[i] = omega * sim->p_f_dot[i]
		                      + (1.0 - omega) * sim->p_f_dot_old[i];

	for (i = 0; i < sim->nz-1; ++i)
		sim->p_f_next_ghost[i + 1] = sim->p_f_dot[i] * sim->dt + sim->p_f_ghost[i + 1];

	set_fluid_bcs(sim->p_f_ghost, sim, p_f_top);
	set_fluid_bcs(sim->p_f_next_ghost, sim, p_f_top);
#ifdef DEBUG
	printf(".. epsilon = %g\n", epsilon);
	puts(".. p_f_dot_expl:");
	print_array(sim->p_f_dot_expl, sim->nz);
	puts(".. p_f_dot_impl:");
	print_array(sim->p_f_dot_impl, sim->nz);
#endif

	for (i = 0; i < sim->nz; ++i)
		if (isnan(sim->p_f_dot_expl[i]) || isinf(sim->p_f_dot_expl[i]))
			errx(1, "invalid: sim->p_f_dot_expl[%d] = %g (t = %g s)",
				 i, sim->p_f_dot_expl[i], sim->t);

	for (i = 0; i < sim->nz; ++i)
		if (isnan(sim->p_f_dot_impl[i]) || isinf(sim->p_f_dot_impl[i]))
			errx(1, "invalid: sim->p_f_dot_impl[%d] = %g (t = %g s)",
				 i, sim->p_f_dot_impl[i], sim->t);

	return solved - 1;
}
