#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <getopt.h>
#include <time.h>
#include <err.h>

#include "simulation.h"
#include "arg.h"

#define EPS 1e-12

/* depth accuracy criteria in meter for solver */
#define TOL 1e-10
#define MAX_ITER 100

/* uncomment to print time spent per time step to stdout */
/* #define BENCHMARK_PERFORMANCE */

#ifndef M_PI
#define M_PI 3.1415926535897932384626433832795028841971693993751058209749445923078164062
#endif

char *argv0;

static void
usage(void)
{
	fprintf(stderr, "usage: %s "
	        "[-a fluid-pressure-ampl] "
	        "[-C fluid-compressibility] "
	        "[-D fluid-diffusivity] "
	        "[-g gravity-accel] "
	        "[-h] "
	        "[-i fluid-viscosity] "
	        "[-k fluid-permeability] "
	        "[-O fluid-pressure-top] "
	        "[-P grain-compressibility] "
	        "[-p grain-porosity] "
	        "[-q fluid-pressure-freq] "
	        "[-R fluid-density] "
	        "[-r grain-density] "
	        "[-v] "
	        "\n", argv0);
	exit(1);
}

static double
skin_depth(const struct simulation *sim)
{
	if (sim->D > 0.0)
		return sqrt(sim->D / (M_PI * sim->p_f_mod_freq));
	else
		return sqrt(sim->k[0] / (sim->mu_f
		                         * (sim->alpha + sim->phi[0] * sim->beta_f)
		                         * M_PI * sim->p_f_mod_freq));
}

/* using alternate form: sin(x) + cos(x) = sqrt(2)*sin(x + pi/4) */
static double
eff_normal_stress_gradient(struct simulation * sim, double d_s, double z_)
{
	if (z_ / d_s > 10.0) {
		fprintf(stderr, "error: %s: unrealistic depth: %g m "
		        "relative to skin depth %g m\n", __func__, z_, d_s);
		free_arrays(sim);
		exit(10);
	}
	return sqrt(2.0) * sin((3.0 * M_PI / 2.0 - z_ / d_s) + M_PI / 4.0)
	       + (sim->rho_s - sim->rho_f) * sim->G * d_s 
	       / sim->p_f_mod_ampl * exp(z_ / d_s);
}

/* use Brent's method for finding roots (Press et al., 2007) */
static double
zbrent(struct simulation *sim,
       double d_s,
       double (*f) (struct simulation *sim, double, double),
       double x1,
       double x2,
       double tol)
{
	int iter;
	double a, b, c, d = NAN, e = NAN, min1, min2,
	       fa, fb, fc, p, q, r, s, tol1, xm;

	a = x1;
	b = x2;
	c = x2;
	fa = (*f) (sim, d_s, a);
	fb = (*f) (sim, d_s, b);

	if ((fa > 0.0 && fb > 0.0) || (fa < 0.0 && fb < 0.0)) {
		warnx("%s: no root in range %g,%g m\n",
		      __func__, x1, x2);
		free_arrays(sim);
		exit(11);
	}
	fc = fb;
	for (iter = 0; iter < MAX_ITER; iter++) {
		if ((fb > 0.0 && fc > 0.0) || (fb < 0.0 && fc < 0.0)) {
			c = a;
			fc = fa;
			d = b - a;
			e = d;
		}
		if (fabs(fc) < fabs(fb)) {
			a = b;
			b = c;
			c = a;
			fa = fb;
			fb = fc;
			fc = fa;
		}
		tol1 = 2.0 * EPS * fabs(b) + 0.5 * tol;
		xm = 0.5 * (c - b);
		if (fabs(xm) <= tol1 || fb == 0.0)
			return b;
		if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
			s = fb / fa;
			if (a == c) {
				p = 2.0 * xm * s;
				q = 1.0 - s;
			} else {
				q = fa / fc;
				r = fb / fc;
				p = s * (2.0 * xm * q * (q - r) - (b - a) * (r - 1.0));
				q = (q - 1.0) * (r - 1.0) * (s - 1.0);
			}
			if (p > 0.0)
				q = -q;
			p = fabs(p);
			min1 = 3.0 * xm * q - fabs(tol1 * q);
			min2 = fabs(e * q);
			if (2.0 * p < (min1 < min2 ? min1 : min2)) {
				e = d;
				d = p / q;
			} else {
				d = xm;
				e = d;
			}
		} else {
			d = xm;
			e = d;
		}
		a = b;
		fa = fb;
		if (fabs(d) > tol1)
			b += d;
		else
			b += ((xm) >= 0.0 ? fabs(tol1) : -fabs(tol1));
		fb = (*f) (sim, d_s, b);
	}
	warnx("error: %s: exceeded maximum number of iterations", __func__);
	free_arrays(sim);
	exit(12);

	return NAN;
}

int
main(int argc, char *argv[])
{
	int i;
	double new_phi, new_k;
	double d_s, depth, depth_limit1, depth_limit2;
	struct simulation sim;

#ifdef BENCHMARK_PERFORMANCE
	clock_t t_begin, t_end;
	double t_elapsed;
#endif

#ifdef __OpenBSD__
	if (pledge("stdio", NULL) == -1)
		err(2, "pledge failed");
#endif

	init_sim(&sim);
	new_phi = sim.phi[0];
	new_k = sim.k[0];

	ARGBEGIN {
	case 'a':
		sim.p_f_mod_ampl = atof(EARGF(usage()));
		break;
	case 'C':
		sim.beta_f = atof(EARGF(usage()));
		break;
	case 'D':
		sim.D = atof(EARGF(usage()));
		break;
	case 'g':
		sim.G = atof(EARGF(usage()));
		break;
	case 'h':
		usage();
		break;
	case 'i':
		sim.mu_f = atof(EARGF(usage()));
		break;
	case 'k':
		new_k = atof(EARGF(usage()));
		break;
	case 'O':
		sim.p_f_top = atof(EARGF(usage()));
		break;
	case 'P':
		sim.alpha = atof(EARGF(usage()));
		break;
	case 'p':
		new_phi = atof(EARGF(usage()));
		break;
	case 'q':
		sim.p_f_mod_freq = atof(EARGF(usage()));
		break;
	case 'R':
		sim.rho_f = atof(EARGF(usage()));
		break;
	case 'r':
		sim.rho_s = atof(EARGF(usage()));
		break;
	case 'v':
		printf("%s-" VERSION "\n", argv0);
		return 0;
		break;
	default:
		usage();
	} ARGEND;

	if (argc > 0)
		usage();

	sim.nz = 2;
	prepare_arrays(&sim);

	if (!isnan(new_phi))
		for (i = 0; i < sim.nz; ++i)
			sim.phi[i] = new_phi;
	if (!isnan(new_k))
		for (i = 0; i < sim.nz; ++i)
			sim.k[i] = new_k;

	check_simulation_parameters(&sim);

	depth = 0.0;
	d_s = skin_depth(&sim);

#ifdef BENCHMARK_PERFORMANCE
	t_begin = clock();
#endif

	/* deep deformatin does not occur with approximately zero
	 * amplitude in water pressure forcing, or if the stress gradient
	 * is positive at zero depth */
	if (fabs(sim.p_f_mod_ampl) > 1e-6 &&
	    eff_normal_stress_gradient(&sim, d_s, 0.0) < 0.0) {

		depth_limit1 = 0.0;
		depth_limit2 = 5.0 * d_s;

		depth = zbrent(&sim,
		               d_s,
		               (double (*) (struct simulation *, double, double))
		               eff_normal_stress_gradient,
		               depth_limit1,
		               depth_limit2,
		               TOL);
	}
#ifdef BENCHMARK_PERFORMANCE
	t_end = clock();
	t_elapsed = (double) (t_end - t_begin) / CLOCKS_PER_SEC;
	printf("time spent = %g s\n", t_elapsed);
#endif

	printf("%.17g\t%.17g\n", depth, d_s);

	free_arrays(&sim);

	return 0;
}
