#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <err.h>
#include "arrays.h"
#include "simulation.h"
#include "fluid.h"


/* iteration limits for solvers */
#define MAX_ITER_GRANULAR 100000
#define MAX_ITER_DARCY 1000000
#define MAX_ITER_STRESS 20000

/* tolerance criteria when in velocity driven or velocity limited mode */
#define RTOL_VELOCITY 1e-3


/* Simulation settings */
void
init_sim(struct simulation *sim)
{
	int ret;

	ret = snprintf(sim->name, sizeof(sim->name), DEFAULT_SIMULATION_NAME);
	if (ret < 0 || (size_t)ret == sizeof(sim->name))
		err(1, "%s: could not write simulation name", __func__);

	sim->G = 9.81;

	sim->P_wall = 120e3;
	sim->mu_wall = 0.45;
	sim->v_x_bot = 0.0;
	sim->v_x_fix = NAN;
	sim->v_x_limit = NAN;

	sim->nz = -1;		/* cell size equals grain size if negative */

	/* lower values of A mean that the velocity curve can have sharper
	 * curves, e.g. at the transition from μ ≈ μ_s */
	sim->A = 0.40;		/* Loose fit to Damsgaard et al 2013 */

	/* lower values of b mean larger shear velocity for a given stress
	 * ratio above yield */
	sim->b = 0.9377;	/* Henann and Kamrin 2016 */

	/* Henann and Kamrin 2016 */
	/* sim->mu_s = 0.3819; */
	/* sim->C = 0.0; */
	/* Damsgaard et al 2013 */
	sim->mu_s = tan(DEG2RAD(22.0));
	sim->C = 0.0;
	sim->phi = initval(0.25, 1);
	sim->d = 0.04;		/* Damsgaard et al 2013 */

	sim->transient = 0;
	sim->phi_min = 0.20;
	sim->phi_max = 0.55;
	sim->dilatancy_constant = 4.09;	/* Pailha & Pouliquen 2009 */

	/* Iverson et al 1997, 1998: Storglaciaren till */
	/* sim->mu_s = tan(DEG2RAD(26.3)); */
	/* sim->C = 5.0e3; */
	/* sim->phi = initval(0.22, 1); */
	/* sim->d = ??; */

	/* Iverson et al 1997, 1998: Two Rivers till */
	/* sim->mu_s = tan(DEG2RAD(17.8)); */
	/* sim->C = 14.0e3; */
	/* sim->phi = initval(0.37, 1); */
	/* sim->d = ??; */

	/* Tulaczyk et al 2000a: Upstream B till */
	/* sim->mu_s = tan(DEG2RAD(23.9)); */
	/* sim->C = 3.0e3; */
	/* sim->phi = initval(0.35, 1); */
	/* sim->d = ??; */

	/* grain material density [kg/m^3] */
	sim->rho_s = 2.6e3;	/* Damsgaard et al 2013 */

	/* spatial settings */
	sim->origo_z = 0.0;
	sim->L_z = 1.0;

	/* temporal settings */
	sim->t = 0.0;
	sim->dt = 1.0;
	sim->t_end = 1.0;
	sim->file_dt = 1.0;
	sim->n_file = 0;

	/* fluid settings */
	sim->fluid = 0;

	sim->rho_f = 1e3;

	/* Water at 20 deg C, Goren et al 2011 */
	/* sim->beta_f = 4.5e-10; */
	/* sim->mu_f = 1.0-3; */

	/* Water at 0 deg C */
	sim->beta_f = 3.9e-10;	/* doi:10.1063/1.1679903 */
	sim->mu_f = 1.787e-3;	/* Cuffey and Paterson 2010 */

	sim->alpha = 1e-8;

	sim->D = -1.0;	/* disabled when negative */

	/* Damsgaard et al 2015 */
	sim->k = initval(1.9e-15, 1);

	/* Iverson et al 1994: Storglaciaren */
	/* sim->k = initval(1.3e-14, 1); */

	/* Engelhardt et al 1990: Upstream B */
	/* sim->k = initval(2.0e-16, 1); */

	/* Leeman et al 2016: Upstream B till */
	/* sim->k = initval(4.9e-17, 1); */

	/* no fluid-pressure variations */
	sim->p_f_top = 0.0;
	sim->p_f_mod_ampl = 0.0;
	sim->p_f_mod_freq = 1.0;
	sim->p_f_mod_phase = 0.0;
	sim->p_f_mod_pulse_time = NAN;
	sim->p_f_mod_pulse_shape = 0;
}

void
prepare_arrays(struct simulation *sim)
{
	if (sim->nz < 2) {
		fprintf(stderr, "error: grid size (nz) must be at least 2 but is %d\n",
		        sim->nz);
		exit(1);
	}
	free(sim->phi);
	free(sim->k);

	sim->z = linspace(sim->origo_z,
	                  sim->origo_z + sim->L_z,
	                  sim->nz);
	sim->dz = sim->z[1] - sim->z[0];
	sim->mu = zeros(sim->nz);
	sim->mu_c = zeros(sim->nz);
	sim->sigma_n_eff = zeros(sim->nz);
	sim->sigma_n = zeros(sim->nz);
	sim->p_f_ghost = zeros(sim->nz + 2);
	sim->p_f_next_ghost = zeros(sim->nz + 2);
	sim->p_f_dot = zeros(sim->nz);
	sim->p_f_dot_expl = zeros(sim->nz);
	sim->p_f_dot_impl = zeros(sim->nz);
	sim->p_f_dot_impl_r_norm = zeros(sim->nz);
	sim->phi = zeros(sim->nz);
	sim->phi_c = zeros(sim->nz);
	sim->phi_dot = zeros(sim->nz);
	sim->k = zeros(sim->nz);
	sim->xi = zeros(sim->nz);
	sim->gamma_dot_p = zeros(sim->nz);
	sim->v_x = zeros(sim->nz);
	sim->d_x = zeros(sim->nz);
	sim->g_local = zeros(sim->nz);
	sim->g_ghost = zeros(sim->nz + 2);
	sim->g_r_norm = zeros(sim->nz);
	sim->I = zeros(sim->nz);
	sim->tan_psi = zeros(sim->nz);
	sim->old_val = empty(sim->nz);
	sim->fluid_old_val = empty(sim->nz);
	sim->tmp_ghost = empty(sim->nz + 2);
	sim->p_f_dot_old = zeros(sim->nz);
}

void
free_arrays(struct simulation *sim)
{
	free(sim->z);
	free(sim->mu);
	free(sim->mu_c);
	free(sim->sigma_n_eff);
	free(sim->sigma_n);
	free(sim->p_f_ghost);
	free(sim->p_f_next_ghost);
	free(sim->p_f_dot);
	free(sim->p_f_dot_expl);
	free(sim->p_f_dot_impl);
	free(sim->p_f_dot_impl_r_norm);
	free(sim->k);
	free(sim->phi);
	free(sim->phi_c);
	free(sim->phi_dot);
	free(sim->xi);
	free(sim->gamma_dot_p);
	free(sim->v_x);
	free(sim->d_x);
	free(sim->g_local);
	free(sim->g_ghost);
	free(sim->g_r_norm);
	free(sim->I);
	free(sim->tan_psi);
	free(sim->old_val);
	free(sim->fluid_old_val);
	free(sim->tmp_ghost);
	free(sim->p_f_dot_old);
}

static void
warn_parameter_value(const char message[],
                     const double value,
                     int *return_status)
{
	fprintf(stderr,
	        "check_simulation_parameters: %s (%.17g)\n",
	        message, value);
	*return_status = 1;
}

static void
check_float(const char name[], const double value, int *return_status)
{
	int ret;
	char message[100];

#ifdef SHOW_PARAMETERS
	printf("%30s: %.17g\n", name, value);
#endif
	if (isnan(value)) {
		ret = snprintf(message, sizeof(message), "%s is NaN", name);
		if (ret < 0 || (size_t)ret >= sizeof(message))
			errx(1, "%s: message parsing", __func__);
		warn_parameter_value(message, value, return_status);
	} else if (isinf(value)) {
		ret = snprintf(message, sizeof(message), "%s is infinite", name);
		if (ret < 0 || (size_t)ret >= sizeof(message))
			errx(1, "%s: message parsing", __func__);
		warn_parameter_value(message, value, return_status);
	}
}

void
check_simulation_parameters(struct simulation *sim)
{
	int return_status = 0;

	check_float("sim->G", sim->G, &return_status);
	if (sim->G < 0.0)
		warn_parameter_value("sim->G is negative", sim->G, &return_status);

	check_float("sim->P_wall", sim->P_wall, &return_status);
	if (sim->P_wall < 0.0)
		warn_parameter_value("sim->P_wall is negative", sim->P_wall,
		                     &return_status);

	check_float("sim->v_x_bot", sim->v_x_bot, &return_status);

	check_float("sim->mu_wall", sim->mu_wall, &return_status);
	if (sim->mu_wall < 0.0)
		warn_parameter_value("sim->mu_wall is negative", sim->mu_wall,
		                     &return_status);

	check_float("sim->A", sim->A, &return_status);
	if (sim->A < 0.0)
		warn_parameter_value("sim->A is negative", sim->A, &return_status);

	check_float("sim->b", sim->b, &return_status);
	if (sim->b < 0.0)
		warn_parameter_value("sim->b is negative", sim->b, &return_status);

	check_float("sim->mu_s", sim->mu_s, &return_status);
	if (sim->mu_s < 0.0)
		warn_parameter_value("sim->mu_s is negative", sim->mu_s,
		                     &return_status);

	check_float("sim->C", sim->C, &return_status);

	check_float("sim->d", sim->d, &return_status);
	if (sim->d <= 0.0)
		warn_parameter_value("sim->d is not a positive number", sim->d,
		                     &return_status);

	check_float("sim->rho_s", sim->rho_s, &return_status);
	if (sim->rho_s <= 0.0)
		warn_parameter_value("sim->rho_s is not a positive number", sim->rho_s,
		                     &return_status);

	if (sim->nz <= 0)
		warn_parameter_value("sim->nz is not a positive number", sim->nz,
		                     &return_status);

	check_float("sim->origo_z", sim->origo_z, &return_status);
	check_float("sim->L_z", sim->L_z, &return_status);
	if (sim->L_z <= sim->origo_z)
		warn_parameter_value("sim->L_z is smaller or equal to sim->origo_z",
		                     sim->L_z, &return_status);

	if (sim->nz <= 0)
		warn_parameter_value("sim->nz is not a positive number", sim->nz,
		                     &return_status);

	check_float("sim->dz", sim->dz, &return_status);
	if (sim->dz <= 0.0)
		warn_parameter_value("sim->dz is not a positive number", sim->dz,
		                     &return_status);

	check_float("sim->t", sim->t, &return_status);
	if (sim->t < 0.0)
		warn_parameter_value("sim->t is a negative number",
		                     sim->t, &return_status);

	check_float("sim->t_end", sim->t_end, &return_status);
	if (sim->t > sim->t_end)
		warn_parameter_value("sim->t_end is smaller than sim->t",
		                     sim->t, &return_status);

	check_float("sim->dt", sim->dt, &return_status);
	if (sim->dt < 0.0)
		warn_parameter_value("sim->dt is less than zero",
		                     sim->dt, &return_status);

	check_float("sim->file_dt", sim->file_dt, &return_status);
	if (sim->file_dt < 0.0)
		warn_parameter_value("sim->file_dt is a negative number",
		                     sim->file_dt, &return_status);

	check_float("sim->phi[0]", sim->phi[0], &return_status);
	if (sim->phi[0] < 0.0 || sim->phi[0] > 1.0)
		warn_parameter_value("sim->phi[0] is not within [0;1]",
		                     sim->phi[0], &return_status);

	check_float("sim->phi_min", sim->phi_min, &return_status);
	if (sim->phi_min < 0.0 || sim->phi_min > 1.0)
		warn_parameter_value("sim->phi_min is not within [0;1]",
		                     sim->phi_min, &return_status);

	check_float("sim->phi_max", sim->phi_max, &return_status);
	if (sim->phi_max < 0.0 || sim->phi_max > 1.0)
		warn_parameter_value("sim->phi_max is not within [0;1]",
		                     sim->phi_max, &return_status);

	check_float("sim->dilatancy_constant", sim->dilatancy_constant, &return_status);
	if (sim->dilatancy_constant < 0.0 || sim->dilatancy_constant > 100.0)
		warn_parameter_value("sim->dilatancy_constant is not within [0;100]",
		                     sim->dilatancy_constant, &return_status);

	if (sim->fluid != 0 && sim->fluid != 1)
		warn_parameter_value("sim->fluid has an invalid value",
		                     (double) sim->fluid, &return_status);

	if (sim->transient != 0 && sim->transient != 1)
		warn_parameter_value("sim->transient has an invalid value",
		                     (double) sim->transient, &return_status);

	if (sim->fluid) {
		check_float("sim->p_f_mod_ampl", sim->p_f_mod_ampl, &return_status);
		if (sim->p_f_mod_ampl < 0.0)
			warn_parameter_value("sim->p_f_mod_ampl is not a zero or positive",
			                     sim->p_f_mod_ampl, &return_status);

		if (sim->P_wall - sim->p_f_mod_ampl < 0.0)
			warn_parameter_value("sim->P_wall - sim->p_f_mod_ampl is negative",
			                     sim->P_wall - sim->p_f_mod_ampl,
			                     &return_status);

		check_float("sim->p_f_mod_freq", sim->p_f_mod_freq, &return_status);
		if (sim->p_f_mod_freq < 0.0)
			warn_parameter_value("sim->p_f_mod_freq is not a zero or positive",
			                     sim->p_f_mod_freq, &return_status);

		check_float("sim->beta_f", sim->beta_f, &return_status);
		if (sim->beta_f <= 0.0)
			warn_parameter_value("sim->beta_f is not positive",
			                     sim->beta_f, &return_status);

		check_float("sim->alpha", sim->alpha, &return_status);
		if (sim->alpha <= 0.0)
			warn_parameter_value("sim->alpha is not positive",
			                     sim->alpha, &return_status);

		check_float("sim->mu_f", sim->mu_f, &return_status);
		if (sim->mu_f <= 0.0)
			warn_parameter_value("sim->mu_f is not positive",
			                     sim->mu_f, &return_status);

		check_float("sim->rho_f", sim->rho_f, &return_status);
		if (sim->rho_f <= 0.0)
			warn_parameter_value("sim->rho_f is not positive",
			                     sim->rho_f, &return_status);

		check_float("sim->k[0]", sim->k[0], &return_status);
		if (sim->k[0] <= 0.0)
			warn_parameter_value("sim->k[0] is not positive",
			                      sim->k[0], &return_status);

		check_float("sim->D", sim->D, &return_status);
		if (sim->transient && sim->D > 0.0)
			warn_parameter_value("constant diffusivity does not work in "
			                     "transient simulations",
			                     sim->D, &return_status);

	}
	if (return_status != 0) {
		fprintf(stderr, "error: aborting due to invalid parameter choices\n");
		exit(return_status);
	}
}

void
lithostatic_pressure_distribution(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->sigma_n[i] = sim->P_wall
		                  + (1.0 - sim->phi[i]) * sim->rho_s * sim->G
		                  * (sim->L_z - sim->z[i]);
}

/* NEW FUNCTIONS START */

inline static double
inertia_number(double gamma_dot_p, double d, double sigma_n_eff, double rho_s)
{
	return gamma_dot_p * d / sqrt(sigma_n_eff / rho_s);
}

void
compute_inertia_number(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->I[i] = inertia_number(sim->gamma_dot_p[i],
		                           sim->d,
		                           sim->sigma_n_eff[i],
		                           sim->rho_s);
}

void
compute_critical_state_porosity(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->phi_c[i] = sim->phi_min
		                + (sim->phi_max - sim->phi_min) * sim->I[i];
}

void
compute_critical_state_friction(struct simulation *sim)
{
	int i;

	if (sim->fluid)
		for (i = 0; i < sim->nz; ++i)
			sim->mu_c[i] = sim->mu_wall / (sim->sigma_n_eff[i]
			                               / (sim->P_wall - sim->p_f_top));
	else
		for (i = 0; i < sim->nz; ++i)
			sim->mu_c[i] = sim->mu_wall
			               / (1.0 + (1.0 - sim->phi[i]) * sim->rho_s * sim->G *
			                  (sim->L_z - sim->z[i]) / sim->P_wall);
}

static void
compute_friction(struct simulation *sim)
{
	int i;

	if (sim->transient)
		for (i = 0; i < sim->nz; ++i)
			sim->mu[i] = sim->mu_c[i] + sim->tan_psi[i];
	else
		for (i = 0; i < sim->nz; ++i)
			sim->mu[i] = sim->mu_c[i];
}

static void
compute_tan_dilatancy_angle(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->tan_psi[i] = sim->dilatancy_constant * (sim->phi_c[i] - sim->phi[i]);
}

static void
compute_porosity_change(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->phi_dot[i] = sim->tan_psi[i] * sim->gamma_dot_p[i] * sim->phi[i];
}

double
kozeny_carman(const double diameter, const double porosity)
{
	return pow(diameter, 2.0) / 180.0
	       * pow(porosity, 3.0)
	       / pow(1.0 - porosity, 2.0);
}

static void
compute_permeability(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->k[i] = kozeny_carman(sim->d, sim->phi[i]);
}

/* NEW FUNCTIONS END */

static double
shear_strain_rate_plastic(const double fluidity, const double friction)
{
	return fluidity * friction;
}

static void
compute_shear_strain_rate_plastic(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->gamma_dot_p[i] = shear_strain_rate_plastic(sim->g_ghost[i + 1],
		                                                sim->mu[i]);
}

static void
compute_shear_velocity(struct simulation *sim)
{
	int i;

	/* TODO: implement iterative solver for v_x from gamma_dot_p field */
	/* Dirichlet BC at bottom */
	sim->v_x[0] = sim->v_x_bot;

	for (i = 1; i < sim->nz; ++i)
		sim->v_x[i] = sim->v_x[i - 1] + sim->gamma_dot_p[i] * sim->dz;
}

void
compute_effective_stress(struct simulation *sim)
{
	int i;

	if (sim->fluid)
		for (i = 0; i < sim->nz; ++i) {
			/* average of current and next step pressure values for effective stress - may not be optimal */
			sim->sigma_n_eff[i] = sim->sigma_n[i]
			                      - ((sim->p_f_ghost[i + 1]
			                          + sim->p_f_next_ghost[i + 1])
			                         / 2.0);
			/* sim->sigma_n_eff[i] = sim->sigma_n[i] - sim->p_f_ghost[i + 1]; */
			if (sim->sigma_n_eff[i] < 0)
				errx(1, "%s: sigma_n_eff[%d] is negative with value %g\n",
				     __func__, i, sim->sigma_n_eff[i]);
		}
	else
		for (i = 0; i < sim->nz; ++i)
			sim->sigma_n_eff[i] = sim->sigma_n[i];
}

static double
cooperativity_length(const double A,
                     const double d,
                     const double mu,
                     const double p,
                     const double mu_s,
                     const double C)
{
	return A * d / sqrt(fabs((mu - C / p) - mu_s));
}

static void
compute_cooperativity_length(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->xi[i] = cooperativity_length(sim->A,
		                                  sim->d,
		                                  sim->mu[i],
		                                  sim->sigma_n_eff[i],
		                                  sim->mu_s,
		                                  sim->C);
}

static double
local_fluidity(const double p,
               const double mu,
               const double mu_s,
               const double C,
               const double b,
               const double rho_s,
               const double d)
{
	if (mu - C / p <= mu_s)
		return 0.0;
	else
		return sqrt(p / rho_s * d * d) * ((mu - C / p) - mu_s) / (b * mu);
}

static void
compute_local_fluidity(struct simulation *sim)
{
	int i;

	for (i = 0; i < sim->nz; ++i)
		sim->g_local[i] = local_fluidity(sim->sigma_n_eff[i],
		                                 sim->mu[i],
		                                 sim->mu_s,
		                                 sim->C,
		                                 sim->b,
		                                 sim->rho_s,
		                                 sim->d);
}

void
set_bc_neumann(double *a,
               const int nz,
               const int boundary,
               const double df,
               const double dx)
{
	if (boundary == -1)
		a[0] = a[1] + df * dx;
	else if (boundary == +1)
		a[nz + 1] = a[nz] - df * dx;
	else
		errx(1, "%s: Unknown boundary %d\n", __func__, boundary);
}

void
set_bc_dirichlet(double *a,
                 const int nz,
                 const int boundary,
                 const double value)
{
	if (boundary == -1)
		a[0] = value;
	else if (boundary == +1)
		a[nz + 1] = value;
	else
		errx(1, "%s: Unknown boundary %d\n", __func__, boundary);
}

double
residual(double new_val, double old_val)
{
	return (new_val - old_val) / (old_val + 1e-16);
}

static void
poisson_solver_1d_cell_update(int i,
                              const double *g_in,
                              const double *g_local,
                              double *g_out,
                              double *r_norm,
                              const double dz,
                              const double *xi)
{
	double coorp_term = dz * dz / (2.0 * pow(xi[i], 2.0));

	g_out[i + 1] = 1.0 / (1.0 + coorp_term)
	               * (coorp_term * g_local[i] + g_in[i + 2] / 2.0
	                  + g_in[i] / 2.0);
	r_norm[i] = fabs(residual(g_out[i + 1], g_in[i + 1]));

#ifdef DEBUG
	printf("-- %d --------------\n", i);
	printf("coorp_term: %g\n", coorp_term);
	printf("   g_local: %g\n", g_local[i]);
	printf("      g_in: %g\n", g_in[i + 1]);
	printf("     g_out: %g\n", g_out[i + 1]);
	printf("    r_norm: %g\n", r_norm[i]);
#endif
}

static int
implicit_1d_jacobian_poisson_solver(struct simulation *sim,
                                    const int max_iter,
                                    const double rel_tol)
{
	int iter, i;
	double r_norm_max = NAN, *tmp;

	for (iter = 0; iter < max_iter; ++iter) {
#ifdef DEBUG
		printf("\n@@@ %s: ITERATION %d @@@\n", __func__, iter);
#endif
		/* Dirichlet BCs resemble fixed particle velocities */
		set_bc_dirichlet(sim->g_ghost, sim->nz, -1, 0.0);
		set_bc_dirichlet(sim->g_ghost, sim->nz, +1, 0.0);

		/* Neumann BCs resemble free surfaces */
		/* set_bc_neumann(sim->g_ghost, sim->nz, +1, 0.0, sim->dz); */

		for (i = 0; i < sim->nz; ++i)
			poisson_solver_1d_cell_update(i,
			                              sim->g_ghost,
			                              sim->g_local,
			                              sim->tmp_ghost,
			                              sim->g_r_norm,
			                              sim->dz,
			                              sim->xi);
		r_norm_max = max(sim->g_r_norm, sim->nz);

		tmp = sim->tmp_ghost;
		sim->tmp_ghost = sim->g_ghost;
		sim->g_ghost = tmp;

		if (r_norm_max <= rel_tol) {
			set_bc_dirichlet(sim->g_ghost, sim->nz, -1, 0.0);
			set_bc_dirichlet(sim->g_ghost, sim->nz, +1, 0.0);
#ifdef DEBUG
			printf(".. Solution converged after %d iterations\n", iter + 1);
#endif
			return 0;
		}
	}

	fprintf(stderr, "implicit_1d_jacobian_poisson_solver: ");
	fprintf(stderr, "Solution did not converge after %d iterations\n", iter + 1);
	fprintf(stderr, ".. Residual normalized error: %f\n", r_norm_max);
	return 1;
}

void
write_output_file(struct simulation *sim, const int normalize)
{
	int ret;
	char outfile[200];
	FILE *fp;

	ret = snprintf(outfile, sizeof(outfile), "%s.output%05d.txt",
	               sim->name, sim->n_file++);
	if (ret < 0 || (size_t)ret >= sizeof(outfile))
		errx(1, "%s: outfile snprintf", __func__);

	if ((fp = fopen(outfile, "w")) != NULL) {
		print_output(sim, fp, normalize);
		fclose(fp);
	} else {
		fprintf(stderr, "could not open output file: %s", outfile);
		exit(1);
	}
}

void
print_output(struct simulation *sim, FILE *fp, const int norm)
{
	int i;
	double *v_x_out;

	if (norm)
		v_x_out = normalize(sim->v_x, sim->nz);
	else
		v_x_out = copy(sim->v_x, sim->nz);

	for (i = 0; i < sim->nz; ++i)
		fprintf(fp,
		        "%.17g\t%.17g\t%.17g\t"
		        "%.17g\t%.17g\t%.17g\t"
		        "%.17g\t%.17g\t%.17g\t%.17g"
		        "\n",
		        sim->z[i],
		        v_x_out[i],
		        sim->sigma_n_eff[i],
		        sim->p_f_ghost[i + 1],
		        sim->mu[i],
		        sim->gamma_dot_p[i],
		        sim->phi[i],
		        sim->I[i],
		        sim->mu[i] * sim->sigma_n_eff[i],
		        sim->d_x[i]);

	free(v_x_out);
}

static void
temporal_increment(struct simulation *sim)
{
	int i;

	if (sim->transient)
		for (i = 0; i < sim->nz; ++i)
			sim->phi[i] += sim->phi_dot[i] * sim->dt;

	if (sim->fluid) {
		for (i = 0; i < sim->nz; ++i) {
			if (isnan(sim->p_f_dot[i])) {
				errx(1, "encountered NaN at sim->p_f_dot[%d] (t = %g s)",
				     i, sim->t);
			} else {
				sim->p_f_ghost[i + 1] += sim->p_f_dot[i] * sim->dt;
			}
		}
	}

	for (i = 0; i < sim->nz; ++i)
		sim->d_x[i] += sim->v_x[i] * sim->dt;

	sim->t += sim->dt;
}

int
coupled_shear_solver(struct simulation *sim,
                     const int max_iter,
                     const double rel_tol)
{
	int i, coupled_iter, stress_iter = 0;
	double r_norm_max, vel_res_norm = NAN, mu_wall_orig = sim->mu_wall;

	copy_values(sim->p_f_ghost, sim->p_f_next_ghost, sim->nz + 2);
	compute_effective_stress(sim);	/* Eq. 9 */

	do {			/* stress iterations */
		coupled_iter = 0;
		do {		/* coupled iterations */

			if (sim->transient) {
				copy_values(sim->phi_dot, sim->old_val, sim->nz);

				/* step 1 */
				compute_inertia_number(sim);	/* Eq. 1 */
				/* step 2 */
				compute_critical_state_porosity(sim);	/* Eq. 2 */
				/* step 3 */
				compute_tan_dilatancy_angle(sim);	/* Eq. 5 */
			}
			compute_critical_state_friction(sim);	/* Eq. 7 */

			/* step 4 */
			if (sim->transient) {
				compute_porosity_change(sim);	/* Eq. 3 */
				compute_permeability(sim);	/* Eq. 6 */
			}
			compute_friction(sim);	/* Eq. 4 */

			/* step 5, Eq. 13 */
			if (sim->fluid && (sim->t > 0) )
				if (darcy_solver_1d(sim, MAX_ITER_DARCY, 1e-3 * rel_tol))
					exit(11);

			/* step 6 */
			compute_effective_stress(sim);	/* Eq. 9 */

			/* step 7 */
			compute_local_fluidity(sim);	/* Eq. 10 */
			compute_cooperativity_length(sim);	/* Eq. 12 */

			/* step 8, Eq. 11 */
			if (implicit_1d_jacobian_poisson_solver(sim,
			                                        MAX_ITER_GRANULAR,
			                                        rel_tol))
				exit(12);

			/* step 9 */
			compute_shear_strain_rate_plastic(sim);	/* Eq. 8 */
			compute_inertia_number(sim);	/* Eq. 1 */
			compute_shear_velocity(sim);

#ifdef DEBUG
			/* for (i = 0; i < sim->nz; ++i) { */
			for (i = sim->nz-1; i < sim->nz; ++i) {
				printf("\nsim->t = %g s\n", sim->t);
				printf("sim->I[%d] = %g\n", i, sim->I[i]);
				printf("sim->phi_c[%d] = %g\n", i, sim->phi_c[i]);
				printf("sim->tan_psi[%d] = %g\n", i, sim->tan_psi[i]);
				printf("sim->mu_c[%d] = %g\n", i, sim->mu_c[i]);
				printf("sim->phi[%d] = %g\n", i, sim->phi[i]);
				printf("sim->phi_dot[%d] = %g\n", i, sim->phi_dot[i]);
				printf("sim->k[%d] = %g\n", i, sim->k[i]);
				printf("sim->mu[%d] = %g\n", i, sim->mu[i]);
			}
#endif

			/* stable porosity change field == coupled solution converged */
			if (sim->transient) {
				for (i = 0; i < sim->nz; ++i)
					sim->g_r_norm[i] = fabs(residual(sim->phi_dot[i], sim->old_val[i]));
				r_norm_max = max(sim->g_r_norm, sim->nz);
				if (r_norm_max <= rel_tol && coupled_iter > 0)
					break;

				if (coupled_iter++ >= max_iter) {
					fprintf(stderr, "coupled_shear_solver: ");
					fprintf(stderr, "Transient solution did not converge "
					        "after %d iterations\n", coupled_iter);
					fprintf(stderr, ".. Residual normalized error: %g\n",
					        r_norm_max);
					return 1;
				}
			}

		} while (sim->transient);
		if (!isnan(sim->v_x_limit) || !isnan(sim->v_x_fix)) {
			if (!isnan(sim->v_x_limit)) {
				vel_res_norm = (sim->v_x_limit - sim->v_x[sim->nz - 1])
				               / (sim->v_x[sim->nz - 1] + 1e-12);
				if (vel_res_norm > 0.0)
					vel_res_norm = 0.0;
			} else {
				vel_res_norm = (sim->v_x_fix - sim->v_x[sim->nz - 1])
				               / (sim->v_x[sim->nz - 1] + 1e-12);
			}
			sim->mu_wall *= 1.0 + (vel_res_norm * 1e-3);
		}
		if (++stress_iter > MAX_ITER_STRESS) {
			fprintf(stderr, "error: stress solution did not converge:\n");
			fprintf(stderr, "v_x=%g, v_x_fix=%g, v_x_limit=%g, "
			        "vel_res_norm=%g, mu_wall=%g\n",
			        sim->v_x[sim->nz - 1], sim->v_x_fix, sim->v_x_limit,
			        vel_res_norm, sim->mu_wall);
			return 10;
		}
	} while ((!isnan(sim->v_x_fix) || !isnan(sim->v_x_limit))
	        && fabs(vel_res_norm) > RTOL_VELOCITY);

	if (!isnan(sim->v_x_limit))
		sim->mu_wall = mu_wall_orig;

	temporal_increment(sim);

	return 0;
}

double
find_flux(const struct simulation *sim)
{
	int i;
	double flux = 0.0;

	for (i = 1; i < sim->nz; ++i)
		flux += (sim->v_x[i] + sim->v_x[i - 1]) / 2.0 * sim->dz;

	return flux;
}

void
set_coupled_fluid_transient_timestep(struct simulation *sim, const double safety)
{
	double max_gamma_dot, mu, xi, max_I, dt;

	/* max expected strain rate */
	max_gamma_dot = 1.0/sim->d;
	if (!isnan(sim->v_x_fix))
		max_gamma_dot = sim->v_x_fix / sim->dz;
	if (!isnan(sim->v_x_limit))
		max_gamma_dot = sim->v_x_limit / sim->dz;

	/* estimate for shear friction */
	mu = (sim->mu_wall / ((sim->sigma_n[sim->nz-1] - sim->p_f_mod_ampl)
	     / (sim->P_wall - sim->p_f_top)))
	     + sim->dilatancy_constant * sim->phi[sim->nz-1];

	/* estimate for cooperativity length */
	xi = cooperativity_length(sim->A,
	                          sim->d,
	                          mu,
	                          (sim->sigma_n[sim->nz - 1] - sim->p_f_mod_ampl),
	                          sim->mu_s,
	                          sim->C);

	/* max expected inertia number */
	max_I = inertia_number(max_gamma_dot,
						   sim->d,
						   sim->sigma_n[sim->nz - 1] - sim->p_f_mod_ampl,
						   sim->rho_s);

	dt = xi * (sim->alpha + sim->phi[sim->nz - 1] * sim->beta_f) * sim->mu_f
	     / (sim->phi[sim->nz - 1] * sim->phi[sim->nz - 1]
	        * sim->phi[sim->nz - 1] * sim->L_z * max_I);

	if (sim->dt > safety * dt)
		sim->dt = safety * dt;
}
